/******************************************************************************
 * Kernel module for MPC74xx performance monitoring counters
 * Filename: pmon.c
 * Feature:  This kernel module is used to read and control all performance 
 *           monitoring counters on MPC7410 and MPC744X processors
 * Note:     Dependent on /dev/pmon char device which is created after insmod
 *           is issued and a major number is shown in /proc/devices. Use 
 *           "mknod /dev/pmon c 254 0"
 *           Once inode is created, user code can then access pmon counters via POSIX
 *           system calls, e.g. open, release, ioctl, read, write.
 *
 * ToDo:     
 * Author:   Jacob Pan
 *
 * History:
 *
 *****************************************************************************/

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <asm/ioctl.h>
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/proc_fs.h>
#include <linux/compatmac.h>
#include "pmon.h"

#if 0
#define MODULE
#define __KERNEL__
#define __REENTRANT__
#endif


static int pmon_read(struct file *file, char* buf, size_t count, loff_t *ppos);

static int pmon_write(struct file *file, const char* buf, size_t count, loff_t *ppos);

static int pmon_open(struct inode *inode, struct file *file);

static int pmon_release(struct inode *inode, struct file *file);

static int pmon_ioctl(struct inode* inode, struct file *file, \
		      unsigned int cmd, unsigned long arg);

static int pmon_proc_read(char* buf, char** start, off_t offset,\
		   int count, int* eof, void* data);


static int major;
/* the ordinary device operations */
static struct file_operations pmon_fops =
{
  owner:   THIS_MODULE,
  read:    pmon_read,
  write:   pmon_write,
  open:    pmon_open,
  release: pmon_release,
  ioctl:   pmon_ioctl
};
/* device open method */
int pmon_open(struct inode *inode, struct file *file)
{
#ifdef LINUX_2_4
  MOD_INC_USE_COUNT;
#endif
  return(0);
}

/* device close method */
int pmon_release(struct inode *inode, struct file *file)
{
#ifdef LINUX_2_4
  MOD_DEC_USE_COUNT;
#endif
  return(0);
}
static int pmon_read(struct file *file, char* buf, size_t count, loff_t *ppos)
{
  unsigned int cycles;
  asm volatile ("mfspr %0, 937" : "=r" (cycles));
  __copy_to_user(buf, &cycles, sizeof(cycles));
  
  return(4);//four bytes 

}
/******************************************************************************
 * Function: pmon_write()
 *           this function is used to write performance monitoring mode control registers
 * Arguments: data passed from application/usr code include selections of PMC events
 *            in char size pointed by char* buf, the user code identifies CPU info to
 *            determine how many PMC counters are available in the CPU. Total PMC number
 *            is passed in count.
 * TODO:     It might be a good idea to check PVR and verify the total PMC counts are 
 *           correct.
 *
 *
 */
static int pmon_write(struct file *file, const char* pmcSel, size_t count, loff_t *ppos)
{
  unsigned int mmcr0=0, mmcr1=0, pmc_default=0;
  
  printk("INFO: calling pmon_write %d bytes\n", count);
  mmcr0 |= ((pmcSel[0] <<(31-25))|(pmcSel[1]));
  mmcr1 |= ((pmcSel[2] <<(31-4)) |(pmcSel[3] <<(31-9)));
  /* clear all counters */
  asm volatile ("mtspr 953, %0" : : "r" (pmc_default));
  asm volatile ("mtspr 954, %0" : : "r" (pmc_default));
  asm volatile ("mtspr 957, %0" : : "r" (pmc_default));
  asm volatile ("mtspr 958, %0" : : "r" (pmc_default));
  if(count == 6)
    {
        mmcr1 |= ((pmcSel[4] <<(31-14))|
                  (pmcSel[5] <<(31-20)));
	asm volatile ("mtspr 945, %0" : : "r" (pmc_default));
	asm volatile ("mtspr 946, %0" : : "r" (pmc_default));

    }
  /* freeze all counters, let usr use ioctl to control start and stop */
  mmcr0 |= 0x80000000;
  asm volatile ("mtspr 952, %0" : : "r" (mmcr0));
  asm volatile ("mtspr 956, %0" : : "r" (mmcr1));

  return(0);
}


static int pmon_ioctl(struct inode* inode, struct file *file, \
		      unsigned int cmd, unsigned long arg)
{

  return(0);

}
static int pmon_proc_read(char* buf, char** start, off_t offset,\
			  int count, int* eof, void* data)
{
  unsigned int upmc1;
  int len=0;
  asm volatile ("mfspr %0, 1017" : "=r" (upmc1));
  len += sprintf(buf+len, "L2CR 0x%08x\n", upmc1);
  *eof =1;
  return (len);
}

/* load the module */
int init_module(void)
{
  struct proc_dir_entry* entry;
  unsigned int l2crVal=0xC0000000;
  mode_t mode =0;
#if 0
  if ((major=register_chrdev(0, "pmon", &pmon_fops))<0) {
    printk("pmon: unable to register character device\n");
    return (-EIO);
  }
#endif

  asm volatile ("mtspr 1017, %0" : : "r" (l2crVal));
  asm volatile ("sync");
  printk("INFO: L2CR set to 0x%08x\n", l2crVal);
  entry = create_proc_read_entry("l2cr",mode ,NULL, pmon_proc_read, NULL);
  if (entry)
    printk(KERN_INFO "created proc entry for l2cr, major =%d\n",major);
  else 
    printk(KERN_INFO "failed proc entry for l2cr\n");
   return (0);
}

/* remove the module */
void cleanup_module(void)
{

  /* unregister the device */
  unregister_chrdev(major, "pmon");
  remove_proc_entry("pmon", NULL);
  printk("removed pmon module\n");
  return;
}

